@@ -62,6 +62,9 @@ static char *hostname_titles[] = {
     N_("Calculator [%s] - Scientific")
 };
 
+/* FIXME: Config value for boolean bitcalculating mode is a string... */
+static char *Rcstr[MAXBITCALC]   = { "NO_BITCALCULATING_MODE", "BITCALCULATING_MODE" };
+
 /*  This table shows the keyboard values that are currently being used:
  *
  *           |  a b c d e f g h i j k l m n o p q r s t u v w x y z
@@ -449,10 +452,10 @@ struct Xobject {               /* Gtk+/Xlib graphics object. */
     GtkWidget *disp[MAXDISPMODES];     /* Numeric display mode. */
     GtkWidget *trig[MAXTRIGMODES];     /* Trigonometric mode. */
 
-    int menuval;                  /* Index to button array at menu time. */
     char *lnp;                    /* Localized numerical point (UTF8 format) */
 
-    int warn_change_mode;    /* Should we warn user when changing modes? */
+    gboolean warn_change_mode;    /* Should we warn user when changing modes? */
+    gboolean bitcalculating_mode;
 };
 
 typedef struct Xobject *XVars;
@@ -468,20 +471,17 @@ enum {
 static XVars X;
 
 
-void
-ui_set_syntax_mode(enum syntax mode)
+/* FIXME: Move this into display.c (reset_display) */
+static void
+reset_display(void)
 {
     struct exprm_state *e;
-    GtkWidget *widget;
 
-    /* TODO: Always do clear things when mode is changed. */      
-    v->syntax = mode;
     switch (v->syntax) {
         case NPA:
             v->noparens = 0;
             MPstr_to_num("0", DEC, v->MPdisp_val);
             show_display(v->MPdisp_val);
-            ui_set_statusbar(_("Activated no operator precedence mode"), "");
             clear_undo_history();
             break;
 
@@ -490,21 +490,36 @@ ui_set_syntax_mode(enum syntax mode)
             MPstr_to_num("0", DEC, e->ans);
             exp_del();
             show_display(e->ans);
-            ui_set_statusbar(
-                 _("Activated expression mode with operator precedence"), "");
             break;
         
         default:
             assert(0);
     }
-    set_resource(R_SYNTAX, Rsstr[v->syntax]);
-    ui_set_mode(v->modetype);
-    // FIXME: We can't allow the display to be editable. See bug #326938
-    gtk_text_view_set_editable(GTK_TEXT_VIEW(X->display_item), 
-                               (v->syntax == EXPRS));
+}
+
+void
+ui_set_syntax_mode(enum syntax mode)
+{
+    GtkWidget *widget;
+
+    v->syntax = mode;
     
-    /* Use loaded Arithmetic Precedence mode setting. */
-    // FIXME: Merge with arithmetic_mode_cb()
+    reset_display();
+    
+    if (mode == NPA) {
+        ui_set_statusbar(_("Activated no operator precedence mode"), "");
+    } else {
+        ui_set_statusbar(
+            _("Activated expression mode with operator precedence"), "");
+    }
+    
+    /* Save the syntax mode */
+    set_resource(R_SYNTAX, Rsstr[v->syntax]);
+
+    /* ??? */
+    ui_set_mode(v->modetype);
+    
+    /* Update menu */
     if (v->syntax == EXPRS) {
         widget = GET_WIDGET("arithmetic_precedence_menu");
         gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
@@ -524,9 +539,9 @@ ui_set_accuracy(int accuracy)
     
     v->accuracy = accuracy;
     
-    SNPRINTF(text, MAXLINE, _("Other (%d) ..."), accuracy);
+    SNPRINTF(text, MAXLINE, _("_Other (%d) ..."), accuracy);
     widget = gtk_bin_get_child(GTK_BIN(GET_WIDGET("acc_item_other")));
-    gtk_label_set_text(GTK_LABEL(widget), text);
+    gtk_label_set_markup_with_mnemonic(GTK_LABEL(widget), text);
 
     desc = g_strdup_printf(ngettext("Set accuracy from 0 to %d numeric places.",
                                     "Set accuracy from 0 to %d numeric places.",
@@ -556,7 +571,7 @@ ui_set_accuracy(int accuracy)
     set_int_resource(R_ACCURACY, accuracy);
     
     ui_make_registers();
-    refresh_display();
+    refresh_display(-1);
     
     /* Hide the manual dialog */
     gtk_widget_hide(X->spframe);
@@ -654,13 +669,24 @@ void
 ui_set_show_bitcalculating(gboolean visible)
 {
     GtkWidget *menu;
-    
-    v->bitcalculating_mode = visible;
+
+    X->bitcalculating_mode = visible;
     ui_set_mode(v->modetype);
-    set_resource(R_BITCALC, Rcstr[v->bitcalculating_mode]);
+    set_resource(R_BITCALC, Rcstr[visible ? 1 : 0]);
 
     menu = GET_WIDGET("show_bitcalculating_menu");
     gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu), visible);
+
+    if (visible) {
+        /* Translators: When the bit calculation extension is visible, by 
+         * clicking on any of these individual bits, their value can be 
+         * toggled, causing the displayed integer value to be adjusted 
+         * accordingly.
+         */
+        ui_set_statusbar(_("Bit calculation extension activated. Click on bit values to toggle them."), "");
+    } else {
+        ui_set_statusbar("", "");
+    }
 }
 
 
@@ -723,9 +749,9 @@ make_hostname()
     *scanner = ':';
     
     if (client_hostname[0] == '\0')
-        return NULL;
+        return (NULL);
     else
-        return(strdup(client_hostname));                
+        return (strdup(client_hostname));                
 }
 
 
@@ -735,59 +761,31 @@ ui_set_mode(enum mode_type mode)
     GtkRequisition *r;
     gint w, h;
     char *hostname, title[MAXLINE];
-    GtkWidget *menu;      
-  
-    switch (mode) {
-        case BASIC:
-            gtk_widget_show(X->bas_panel);
-            gtk_widget_hide(X->adv_panel);
-            gtk_widget_hide(X->fin_panel);
-            gtk_widget_hide(X->mode_panel);
-            gtk_widget_hide(X->bit_panel);
-            gtk_widget_hide(X->sci_panel);
-            menu = GET_WIDGET("view_basic_menu");
-            break;
+    GtkWidget *menu;
 
-        case ADVANCED:
-            gtk_widget_hide(X->bas_panel);
-            gtk_widget_show(X->adv_panel);
-            gtk_widget_hide(X->fin_panel);
-            gtk_widget_hide(X->mode_panel);
-            gtk_widget_hide(X->bit_panel);
-            gtk_widget_hide(X->sci_panel);
-            menu = GET_WIDGET("view_advanced_menu");
-            break;
-
-        case FINANCIAL:
-            gtk_widget_hide(X->bas_panel);
-            gtk_widget_show(X->adv_panel);
-            gtk_widget_show(X->fin_panel);
-            gtk_widget_hide(X->mode_panel);
-            gtk_widget_hide(X->bit_panel);
-            gtk_widget_hide(X->sci_panel);
-            menu = GET_WIDGET("view_financial_menu");
-            break;
-
-        case SCIENTIFIC:
-            gtk_widget_hide(X->bas_panel);
-            gtk_widget_show(X->adv_panel);
-            gtk_widget_hide(X->fin_panel);
-            gtk_widget_show_all(X->mode_panel);
-            if (v->bitcalculating_mode) {
-                gtk_widget_show_all(X->bit_panel);
-            } else {
-                gtk_widget_hide(X->bit_panel);
-            }
-            gtk_widget_show(X->sci_panel);
-            menu = GET_WIDGET("view_scientific_menu");
-            break;
-        
-        default:
-            assert(FALSE);
-            return;
-    }
-   
+    refresh_display(-1);
+    ui_make_registers();
     
+    /* Save mode */
+    set_resource(R_MODE, Rmstr[(int) v->modetype]);
+    
+    /* Show/enable the widgets used in this mode */
+    g_object_set(G_OBJECT(X->bas_panel),  "visible", mode == BASIC, NULL);
+    g_object_set(G_OBJECT(X->adv_panel),  "visible", mode != BASIC, NULL);
+    g_object_set(G_OBJECT(X->fin_panel),  "visible", mode == FINANCIAL, NULL);
+    g_object_set(G_OBJECT(X->mode_panel), "visible", mode == SCIENTIFIC, NULL);
+    g_object_set(G_OBJECT(X->sci_panel),  "visible", mode == SCIENTIFIC, NULL);
+    g_object_set(G_OBJECT(X->bit_panel),  "visible",
+                 mode == SCIENTIFIC && X->bitcalculating_mode, NULL);
+    gtk_widget_set_sensitive(GET_WIDGET("show_bitcalculating_menu"),
+                             mode == SCIENTIFIC);
+    gtk_widget_set_sensitive(GET_WIDGET("show_trailing_zeroes_menu"),
+                             mode == SCIENTIFIC);
+    gtk_widget_set_sensitive(GET_WIDGET("show_registers_menu"),
+                             mode != BASIC);
+    
+    /* HACK: Some horrible hack down below to keep the buttons the same size.
+     * There must be a safer way of doing this... */
     r = g_new0(GtkRequisition, 1);
     gtk_widget_size_request(X->menubar, r);
     w = r->width;
@@ -813,13 +811,13 @@ ui_set_mode(enum mode_type mode)
         w = MAX(w, r->width);
         h += r->height;
     }
+    g_free(r);
   
     /* For initial display. */
     gtk_window_set_default_size(GTK_WINDOW(X->kframe), w, h);
     gtk_window_resize(GTK_WINDOW(X->kframe), w, h);
-  
-    g_free(r);
-    
+
+    /* Set the title */
     if((hostname = make_hostname())) {
         SNPRINTF(title, MAXLINE, hostname_titles[mode], hostname);
         g_free(hostname);
@@ -828,6 +826,28 @@ ui_set_mode(enum mode_type mode)
     }
     gtk_window_set_title(GTK_WINDOW(X->kframe), title);
 
+    /* Update the menu */
+    switch (mode) {
+        case BASIC:
+            menu = GET_WIDGET("view_basic_menu");
+            break;
+
+        case ADVANCED:
+            menu = GET_WIDGET("view_advanced_menu");
+            break;
+
+        case FINANCIAL:
+            menu = GET_WIDGET("view_financial_menu");
+            break;
+
+        case SCIENTIFIC:
+            menu = GET_WIDGET("view_scientific_menu");
+            break;
+        
+        default:
+            assert(FALSE);
+            return;
+    }
     gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu), TRUE);
 }
 
@@ -956,9 +976,10 @@ scroll_right()
 
 
 void
-ui_set_display(char *str)
+ui_set_display(char *str, int cursor)
 {
     char localized[MAX_LOCALIZED];
+    GtkTextIter iter;
 
     if (str == NULL || str[0] == '\0') {
         str = " ";
@@ -968,8 +989,19 @@ ui_set_display(char *str)
             str = localized;
         }
     }
+    
     gtk_text_buffer_set_text(X->display_buffer, str, -1);
-    scroll_right();
+    
+    if (cursor < 0) {
+        gtk_text_buffer_get_end_iter(X->display_buffer, &iter);
+    } else {        
+        gtk_text_buffer_get_iter_at_offset(X->display_buffer, &iter, cursor);
+    }
+    gtk_text_buffer_place_cursor(X->display_buffer, &iter);
+    
+    if (cursor < 0) {
+        scroll_right();
+    }
 }
 
 
@@ -1036,7 +1068,7 @@ ui_get_localized_numeric_point(void)
 
     decimal_point = localeconv()->decimal_point;
 
-    return(g_locale_to_utf8(decimal_point, -1, NULL, NULL, NULL));
+    return (g_locale_to_utf8(decimal_point, -1, NULL, NULL, NULL));
 }
 
 
@@ -1058,6 +1090,7 @@ void
 ui_set_registers_visible(gboolean visible)
 {
     GtkWidget *menu;
+
     ui_make_registers();
 
     menu = GET_WIDGET("show_registers_menu");
@@ -1084,20 +1117,26 @@ gchar *
 ui_get_display(void)
 {
     GtkTextIter start, end;
-    gtk_text_buffer_get_bounds(X->display_buffer, &start, &end);   
-    return gtk_text_buffer_get_text(X->display_buffer,
-                                    &start,
-                                    &end,
-                                    FALSE);
+    gtk_text_buffer_get_bounds(X->display_buffer, &start, &end);
+    return (gtk_text_buffer_get_text(X->display_buffer,
+                                     &start,
+                                     &end,
+                                     FALSE));
 }
 
 
-int
-ui_get_cursor(void)
+static int
+get_cursor(void)
 {
     gint pos;
     g_object_get(G_OBJECT(X->display_buffer), "cursor-position", &pos, NULL);
-    return pos;
+    
+    /* Convert the last position to -1 */
+    if (pos == gtk_text_buffer_get_char_count(X->display_buffer)) {
+        return (-1);
+    } else {
+        return (pos);
+    }
 }
 
 
@@ -1234,7 +1273,7 @@ static gboolean
 aframe_delete_cb(GtkWidget *dialog)
 {
     aframe_response_cb(dialog, GTK_RESPONSE_CANCEL);
-    return(TRUE);
+    return (TRUE);
 }
 
 
@@ -1258,7 +1297,7 @@ static gboolean
 rframe_delete_cb(GtkWidget *dialog)
 {
     rframe_response_cb(dialog, GTK_RESPONSE_OK);
-    return(TRUE);
+    return (TRUE);
 }
 
 
@@ -1287,24 +1326,18 @@ base_cb(GtkWidget *widget)
 
 static void do_button(int function, int arg)
 {
-    struct exprm_state *e;
-
     switch (v->syntax) {
         case NPA:
             process_item(&buttons[function], arg);
             set_bit_panel();
             if (v->new_input && v->dtype == FIX) {
                 STRNCPY(v->fnum, v->display, MAX_DIGITS - 1);
-                ui_set_display(v->fnum);
+                ui_set_display(v->fnum, -1);
             }
             break;
 
         case EXPRS:
-            e = get_state();
-            e->value = arg;
-            MEMCPY(&(e->button), &buttons[function], sizeof(struct button));
-            new_state();
-            do_expression();
+            do_expression(function, arg, get_cursor());
             set_bit_panel();
             break;
 
@@ -1322,7 +1355,6 @@ help_display(void)
     const char *lang;
     char *uri = NULL;
     GdkScreen *gscreen;
-
     int i;
     
     const char * const * langs = g_get_language_names ();
@@ -1512,7 +1544,7 @@ static gboolean
 edit_constants_delete_cb(GtkDialog *dialog)
 {
     edit_constants_response_cb(dialog, GTK_RESPONSE_CANCEL);
-    return TRUE;
+    return (TRUE);
 }
 
 
@@ -1550,7 +1582,7 @@ static gboolean
 edit_functions_delete_cb(GtkDialog *dialog)
 {
     edit_functions_response_cb(dialog, GTK_RESPONSE_CANCEL);
-    return TRUE;
+    return (TRUE);
 }
 
 
@@ -1573,7 +1605,7 @@ create_constants_model()
                            -1);
     }
 
-    return(GTK_TREE_MODEL(model));
+    return (GTK_TREE_MODEL(model));
 }
 
 
@@ -1596,7 +1628,7 @@ create_functions_model()
                            -1);
     }
 
-    return(GTK_TREE_MODEL(model));
+    return (GTK_TREE_MODEL(model));
 }
 
 
@@ -1616,25 +1648,6 @@ ui_make_registers()            /* Calculate memory register frame values. */
 }
 
 
-static void
-reset_mode_display(void)
-{
-    switch (v->syntax) {
-        case NPA:
-            show_display(v->MPdisp_val);
-            break;
-
-        case EXPRS:
-            refresh_display();
-            break;
-        
-        default:
-            assert(FALSE);
-    }
-    ui_make_registers();
-}
-
-
 static void
 save_win_position()
 {
@@ -1653,12 +1666,13 @@ change_mode(int mode)
 
     ui_set_base(DEC);
     ui_set_numeric_mode(FIX);
-    ui_set_accuracy(9);
+    ui_set_accuracy(DEFAULT_ACCURACY);
     ui_set_show_thousands_seperator(FALSE);
     ui_set_show_trailing_zeroes(FALSE);
+    ui_set_mode(v->modetype);
 
-    reset_mode_display();
-    do_mode(TRUE);
+    /* Reset display */
+    do_clear();
 }
 
 
@@ -1668,8 +1682,9 @@ request_change_mode()
     GtkWidget *dialog, *request_check, *button;
     gint response;
     
-    if (!X->warn_change_mode)
-        return TRUE;
+    if (!X->warn_change_mode) {
+        return (TRUE);
+    }
 
     dialog = gtk_message_dialog_new(GTK_WINDOW(X->kframe),
                           GTK_DIALOG_MODAL|GTK_DIALOG_DESTROY_WITH_PARENT,
@@ -1712,7 +1727,7 @@ request_change_mode()
     
     gtk_widget_destroy(dialog);
 
-    return response == GTK_RESPONSE_ACCEPT;
+    return (response == GTK_RESPONSE_ACCEPT);
 }
 
 
@@ -1723,11 +1738,26 @@ bit_toggle_cb(GtkWidget *event_box, GdkEventButton *event)
     double number;
     unsigned long long lval;
     int n, MP1[MP_SIZE], index;
+    struct exprm_state *e =  get_state();
 
     index = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(event_box),
                                               "bit_index"));
     n = MAXBITS - index - 1;
-    MPstr_to_num(v->display, v->base, MP1);
+
+    switch (v->syntax) {
+        case NPA:
+            MPstr_to_num(v->display, v->base, MP1);
+            break;
+        case EXPRS: {
+            int ret = usable_num(e->ans);
+            assert(!ret);
+            mpstr(e->ans, MP1);
+        }
+        break;
+        default:
+            assert(FALSE);
+    }
+
     mpcmd(MP1, &number);
     lval = (long long) number;
 
@@ -1738,13 +1768,24 @@ bit_toggle_cb(GtkWidget *event_box, GdkEventButton *event)
         lval |=  (1LL << n);
         gtk_label_set_text(GTK_LABEL(X->bits[index]), " 1");
     }
-
     number = (double) lval;
-    mpcdm(&number, v->MPdisp_val);
-    show_display(v->MPdisp_val);
-    v->toclear = 0;
 
-    return(TRUE);
+    switch (v->syntax) {
+        case NPA:
+            mpcdm(&number, v->MPdisp_val);
+            show_display(v->MPdisp_val);
+            break;
+        case EXPRS:
+            mpcdm(&number, e->ans);
+            exp_replace("Ans");
+            refresh_display(-1);
+            break;
+        default:
+            assert(FALSE);
+    }
+
+    v->toclear = 0;
+    return (TRUE);
 }
 
 
@@ -1882,7 +1923,7 @@ check_for_localized_numeric_point(int keyval)
 
     outbuf[g_unichar_to_utf8(ch, outbuf)] = '\0';
 
-    return(strcmp(outbuf, X->lnp) == 0);
+    return (strcmp(outbuf, X->lnp) == 0);
 }
 
 
@@ -1910,26 +1951,6 @@ inv_cb(GtkWidget *widget)
 }
 
 
-/*ARGSUSED*/
-static gboolean
-display_focus_out_cb(GtkWidget *widget, GdkEventKey *event)
-{
-    if (v->syntax == EXPRS) {
-        exp_replace(ui_get_display());
-    }
-    return(FALSE);
-}
-
-
-/*ARGSUSED*/
-static gboolean
-display_focus_in_cb(GtkWidget *widget, GdkEventKey *event)
-{
-    v->ghost_zero = 0;
-    return(FALSE);
-}
-
-
 /*ARGSUSED*/
 static void
 menu_pos_func(GtkMenu *menu, gint *x, gint *y,
@@ -2000,7 +2021,7 @@ kframe_key_press_cb(GtkWidget *widget, GdkEventKey *event)
 {
     int i, j, state;
     GtkWidget *button;
-
+    
     if (check_for_localized_numeric_point(event->keyval) == TRUE) {
         event->state = 0;
         event->keyval = GDK_KP_Decimal;
@@ -2045,12 +2066,14 @@ kframe_key_press_cb(GtkWidget *widget, GdkEventKey *event)
     }
     
     /* Connect home and end keys to move into the display entry */
-    if (event->keyval == GDK_Home) {
-        select_display_entry(0);
-        return (TRUE);
-    } else if (event->keyval == GDK_End) {
-        select_display_entry(-1);
-        return (TRUE);
+    if (!gtk_widget_is_focus(X->display_item)) {
+        if (event->keyval == GDK_Home) { /* || event->keyval == GDK_Left) { */
+            select_display_entry(0);
+            return (TRUE);
+        } else if (event->keyval == GDK_End) { /* || event->keyval == GDK_Right) { */
+            select_display_entry(-1);
+            return (TRUE);
+        }
     }
 
     for (i = 0; i < NBUTTONS; i++) {
@@ -2067,13 +2090,13 @@ kframe_key_press_cb(GtkWidget *widget, GdkEventKey *event)
             if (button_widgets[i].accelerator_keys[j] == event->keyval &&
                 (button_widgets[i].accelerator_mods[j] & ~GDK_SHIFT_MASK) == state) {
                 button_cb(button, NULL);
-                return(TRUE);
+                return (TRUE);
             }
             j++;
         }
     }
 
-    return(FALSE);
+    return (FALSE);
 }
 
 
@@ -2105,7 +2128,7 @@ static void
 get_proc(GtkClipboard *clipboard, const gchar *buffer, gpointer data)
 {
     int ret;
-    gchar *dstp, *end_buffer, *srcp, *text;
+    gchar *dstp, *end_buffer, *srcp, *text, c;
 
     if (buffer == NULL) {
         return;
@@ -2114,50 +2137,61 @@ get_proc(GtkClipboard *clipboard, const gchar *buffer, gpointer data)
     end_buffer = (gchar *) (buffer + strlen(buffer));
     text = malloc(strlen(buffer)+1);
 
-    srcp = (gchar *) buffer;
     dstp = text;
-    while (srcp < end_buffer) {
-
+    for (srcp = (gchar *) buffer; srcp < end_buffer; srcp++) {
         /* If the clipboard buffer contains any occurances of the "thousands
          * separator", remove them.
          */
         if (*srcp == v->tsep[0]) {
             if (strstr(srcp, v->tsep) == srcp) {
-                srcp += strlen(v->tsep);
-            } else {
-                *dstp++ = *srcp++;
-            }
-        } else {
-
-            /* If an "A", "B", "C", "D" or "F" character is encountered, it 
-             * will be converted to its lowercase equivalent. If an "E" is 
-             * found,  and the next character is a "-" or a "+", then it 
-             * remains as an upper case "E" (it's assumed to be a possible 
-             * exponential number), otherwise its converted to a lower case 
-             * "e". See bugs #455889 and #469245 for more details.
-             */
-            switch (*srcp) {
-                case 'A':
-                case 'B':
-                case 'C':
-                case 'D':
-                case 'F': *dstp++ = tolower(*srcp);
-                          srcp++;
-                          break;
-
-                case 'E': if (srcp < (end_buffer-1)) {
-                              if (*(srcp+1) != '-' &&
-                                  *(srcp+1) != '+') {
-                                  *dstp++ = tolower(*srcp);
-                                  srcp++;
-                                  break;
-                              }
-                          }
-                          /*FALLTHROUGH*/
-
-                default:  *dstp++ = *srcp++;
+                srcp += strlen(v->tsep) - 1;
+                continue;
             }
         }
+
+        /* If an "A", "B", "C", "D" or "F" character is encountered, it 
+         * will be converted to its lowercase equivalent. If an "E" is 
+         * found,  and the next character is a "-" or a "+", then it 
+         * remains as an upper case "E" (it's assumed to be a possible 
+         * exponential number), otherwise its converted to a lower case 
+         * "e". See bugs #455889 and #469245 for more details.
+         */
+        switch (*srcp) {
+            /* Replace tabs with spaces */
+            case '\t':
+                c = ' ';
+                break;
+                
+            /* Terminate on newlines */
+            case '\r':
+            case '\n':
+                c = '\0';
+                break;
+                
+            case 'A':
+            case 'B':
+            case 'C':
+            case 'D':
+            case 'F':
+                c = tolower(*srcp);
+                break;
+
+            case 'E':
+                c = *srcp;
+                if (srcp < (end_buffer-1)) {
+                    if (*(srcp+1) != '-' &&
+                        *(srcp+1) != '+') {
+                        c = tolower(*srcp);
+                    }
+                }
+                break;
+            
+            default:
+                c = *srcp;
+                break;
+        }
+        
+        *dstp++ = c;
     }
     *dstp++ = '\0';
 
@@ -2173,8 +2207,8 @@ get_proc(GtkClipboard *clipboard, const gchar *buffer, gpointer data)
             break;
     
         case EXPRS:
-            exp_append((char *) text);
-            refresh_display();
+            exp_insert((char *) text, get_cursor()); // FIXME: Move out of gtk.c
+            refresh_display(-1);
             break;
     
         default:
@@ -2193,7 +2227,7 @@ mouse_button_cb(GtkWidget *widget, GdkEventButton *event)
                                    get_proc, NULL);
     }
 
-    return(FALSE);
+    return (FALSE);
 }
 
 
@@ -2219,7 +2253,6 @@ static void
 undo_cb(GtkWidget *widget)
 {
     perform_undo();
-    refresh_display();
 }
 
 
@@ -2228,11 +2261,10 @@ static void
 redo_cb(GtkWidget *widget)
 {
     perform_redo();
-    refresh_display();
+    refresh_display(-1);
 }
 
 
-//FIXME
 /*ARGSUSED*/
 static void
 for_each_menu(GtkWidget *widget, gpointer data)
@@ -2315,10 +2347,7 @@ arithmetic_mode_cb(GtkWidget *widget)
 static void
 mode_radio_cb(GtkWidget *menu)
 {
-    struct exprm_state *e;
     int mode;             /* The new mode. */
-    int immediate = 0;    /* Set if we can change mode without warning user. */
-    int complete = 0;     /* Set if the user has completed a calculation. */
 
     if (!gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(menu))) {
         return;
@@ -2339,39 +2368,15 @@ mode_radio_cb(GtkWidget *menu)
  *
  * (unless we are in Scientific mode with Decimal numeric base and Fixed).
  */
-
-    switch (v->syntax) {
-        case NPA:
-            if (v->old_cal_value < 0 ||
-                v->old_cal_value == KEY_CALCULATE) {
-                complete = 1;    /* Calculation is complete. */
-            }
-            break;
-
-        case EXPRS:
-            e = get_state();
-            if (!e->expression || !strcmp(e->expression, "Ans")) {
-                complete = 1;   /* Calculation is complete. */
-            }
-            break;
-        
-        default:
-            assert(FALSE);
-    }
-
-    if (complete) {
-        if ((v->modetype != SCIENTIFIC) ||
-            (v->dtype == FIX && v->base == DEC)) {
-            immediate = 1;
-        }
-    }
-
-    if (immediate) {
+    if (display_is_result() &&
+        ((v->modetype != SCIENTIFIC) ||
+         (v->dtype == FIX && v->base == DEC))) {
         v->modetype = mode;
-        reset_mode_display();
-        do_mode(FALSE);
-    } else if (request_change_mode()) {
-        change_mode(mode);
+        ui_set_mode(v->modetype);
+    } else {
+        if (request_change_mode()) {
+            change_mode(mode);
+        }
     }
 }
 
@@ -2400,6 +2405,14 @@ accuracy_other_cb(GtkWidget *widget)
 }
 
 
+/*ARGSUSED*/
+static void
+accuracy_default_cb(GtkWidget *widget)
+{
+    do_accuracy(DEFAULT_ACCURACY);
+}
+
+
 /*ARGSUSED*/
 static void
 show_trailing_zeroes_cb(GtkWidget *widget)
@@ -2437,7 +2450,7 @@ static gboolean
 spframe_delete_cb(GtkWidget *dialog)
 {
     spframe_response_cb(dialog, GTK_RESPONSE_CANCEL);
-    return(TRUE);
+    return (TRUE);
 }
 
 
@@ -2572,6 +2585,7 @@ create_kframe()
     CONNECT_SIGNAL(show_registers_cb);
     CONNECT_SIGNAL(accuracy_radio_cb);
     CONNECT_SIGNAL(accuracy_other_cb);
+    CONNECT_SIGNAL(accuracy_default_cb);    
     CONNECT_SIGNAL(constant_menu_cb);
     CONNECT_SIGNAL(function_menu_cb);
     CONNECT_SIGNAL(store_menu_cb);
@@ -2579,8 +2593,6 @@ create_kframe()
     CONNECT_SIGNAL(exchange_menu_cb);
     CONNECT_SIGNAL(arithmetic_mode_cb);
     CONNECT_SIGNAL(mouse_button_cb);
-    CONNECT_SIGNAL(display_focus_in_cb);
-    CONNECT_SIGNAL(display_focus_out_cb);
     /* Detect when populating the right-click menu to enable pasting */
     CONNECT_SIGNAL(buffer_populate_popup_cb);
     CONNECT_SIGNAL(shift_cb);
@@ -2787,7 +2799,6 @@ create_kframe()
     add_cf_column(GTK_TREE_VIEW(treeview), _("Description"),
                   COLUMN_DESCRIPTION, TRUE);
 
-
     X->display_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(X->display_item));
     gtk_widget_ensure_style(X->display_item);
     font_desc = pango_font_description_copy(X->display_item->style->font_desc);
@@ -2795,9 +2806,6 @@ create_kframe()
     gtk_widget_modify_font(X->display_item, font_desc);
     pango_font_description_free(font_desc);
     gtk_widget_set_name(X->display_item, "displayitem");
-    // FIXME: We can't allow the display to be editable. See bug #326938
-    gtk_text_view_set_editable(GTK_TEXT_VIEW(X->display_item), 
-                               (v->syntax == EXPRS));
     atk_object_set_role(gtk_widget_get_accessible(X->display_item), 
                                                   ATK_ROLE_EDITBAR);
 
@@ -2886,18 +2894,24 @@ void
 ui_load()
 {
     int boolval;
+    char *resource, text[MAXLINE];
+    gboolean show_bit;
+    GtkWidget *widget;
 
     read_cfdefs();
     
     /* Create main gcalctool window. */
     create_kframe();
+    
+    /* Load configuration */
+    resource = get_resource(R_BITCALC);
+    show_bit = resource != NULL && strcmp(resource, Rcstr[0]) != 0;
 
     ui_set_show_thousands_seperator(v->show_tsep);
     ui_set_show_trailing_zeroes(v->show_zeroes);
-    ui_set_show_bitcalculating(v->bitcalculating_mode);
+    ui_set_show_bitcalculating(show_bit);
     
     ui_set_syntax_mode(v->syntax);
-    ui_set_display("0.00");
     ui_set_mode(v->modetype);
     ui_set_numeric_mode(FIX);
     ui_set_base(v->base);
@@ -2918,13 +2932,16 @@ ui_load()
         gtk_window_set_focus(GTK_WINDOW(X->kframe),
                              GTK_WIDGET(X->clear_buttons[1]));
     }
+
+    /* Set default accuracy menu item */
+    SNPRINTF(text, MAXLINE, _("Reset to _Default (%d)"), DEFAULT_ACCURACY);
+    widget = gtk_bin_get_child(GTK_BIN(GET_WIDGET("acc_item_default")));
+    gtk_label_set_markup_with_mnemonic(GTK_LABEL(widget), text);
 }
 
 void
 ui_start()
 {
-    struct exprm_state *e;
-    
     X->warn_change_mode = TRUE; // FIXME: Load from GConf
     
     ui_set_base(v->base);
@@ -2933,24 +2950,10 @@ ui_start()
 
     gtk_widget_show(X->kframe);
 
-    switch (v->syntax) { 
-        case NPA:
-            break;
-
-        case EXPRS:
-            /* Init expression mode.
-             * This must be executed after do_base is called at init.
-             * FIXME: The init code here is duplicated elsewhere.
-             */
-            e = get_state();
-            MPstr_to_num("0", DEC, e->ans);
-            exp_del();
-            show_display(e->ans);
-            break;
-
-        default:
-            assert(0);
-    }
+    /* Init expression mode.
+     * This must be executed after do_base is called at init.
+     */    
+    reset_display();
 
     gtk_main();
 }
