@@ -1,7 +1,7 @@
 
-/*  $Header: /cvs/gnome/gcalctool/gcalctool/lr_parser.c,v 1.7 2006/01/09 16:30:20 richb Exp $
+/*  $Header$
  *
- *  Copyright (C) 2004-2006 Sami Pietila
+ *  Copyright (C) 2004-2007 Sami Pietila
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,8 +21,8 @@
 
 #include "lr_parser.h"
 #include "limits.h"
-#include "extern.h"
 #include "calctool.h"
+#include "functions.h" /* FIXME: Needed for gc_strdup() */
 
 /* TODO: This file is almost identical to ce-parser. */
 
